# 4 - Designing a Simple Text Editor

This chapter takes you through the design of a simple pyFLTK-based text editor.

## Determining the Goals of the Text Editor

Since this will be the first big project you'll be doing with pyFLTK, lets define what we want our text editor to do:

1. Provide a menubar/menus for all functions.
2. Edit a single text file, possibly with multiple views.
3. Load from a file.
4. Save to a file.
5. Cut/copy/delete/paste functions.
6. Search and replace functions.
7. Keep track of when the file has been changed.

## Designing the Main Window

Now that we've outlined the goals for our editor, we can begin with the design of our GUI. Obviously the first thing that we need is a window, which we'll place inside a class called `EditorWindow`:

    class EditorWindow(Fl_Double_Window):
        search = ""
        def __init__(self, w, h, label):
            Fl_Double_Window.__init__(self, w, h, label)

## Variables

Our text editor will need some global variables to keep track of things:

    changed = False
    filename = ""
    title = ""
    textbuf = None
    editor = None

The `textbuf` variable is the text editor buffer for our window class described previously. We'll cover the other variables as we build the application.

## Menubars and Menus

The first goal requires us to use a menubar and menus that define each function the editor needs to perform. The `Fl_Menu_Item` structure is used to define the menus and items in a menubar:

    menuitems = (( "&File",              0, 0, 0, FL_SUBMENU ),
        ( "&New File",        0, new_cb ),
        ( "&Open File...",    FL_CTRL + ord('o'), open_cb ),
        ( "&Insert File...",  FL_CTRL + ord('i'), insert_cb, 0, FL_MENU_DIVIDER ),
        ( "&Save File",       FL_CTRL + ord('s'), save_cb ),
        ( "Save File &As...", FL_CTRL + FL_SHIFT + ord('s'), saveas_cb, 0, FL_MENU_DIVIDER ),
        ( "New &View", FL_ALT + ord('v'), view_cb, 0 ),
        ( "&Close View", FL_CTRL + ord('w'), close_cb, 0, FL_MENU_DIVIDER ),
        ( "E&xit", FL_CTRL + ord('q'), quit_cb, 0 ),
        ( None, 0 ),

      ( "&Edit", 0, 0, 0, FL_SUBMENU ),
        ( "Cu&t",        FL_CTRL + ord('x'), cut_cb ),
        ( "&Copy",       FL_CTRL + ord('c'), copy_cb ),
        ( "&Paste",      FL_CTRL + ord('v'), paste_cb ),
        ( "&Delete",     0, delete_cb ),
        ( None, 0 ),

      ( "&Search", 0, 0, 0, FL_SUBMENU ),
        ( "&Find...",       FL_CTRL + ord('f'), find_cb ),
        ( "F&ind Again",    FL_CTRL + ord('g'), find2_cb ),
        ( "&Replace...",    FL_CTRL + ord('r'), replace_cb ),
        ( "Re&place Again", FL_CTRL + ord('t'), replace2_cb ),
        ( None, 0 ),

      ( None, 0 )
    )

Once we have the menus defined we can create the `Fl_Menu_Bar` widget and assign the menus to it with:

    m = Fl_Menu_Bar(0, 0, 660, 30);
    m.copy(menuitems)

We'll define the callback functions later.

## Editing the Text

To keep things simple our text editor will use the `Fl_Text_Editor` widget to edit the text:

    w.editor = Fl_Text_Editor(0, 30, 660, 370)
    w.editor.buffer(textbuf)

So that we can keep track of changes to the file, we also want to add a "modify" callback:

    textbuf.add_modify_callback(changed_cb, w)
    textbuf.call_modify_callbacks()

Finally, we want to use a mono-spaced font like `FL_COURIER`:

    w.editor.textfont(FL_COURIER)

## The Replace Dialog

We can use the pyFLTK convenience functions for many of the editor's dialogs, however the replace dialog needs its own custom window. To keep things simple we will have a "find" string, a "replace" string, and "replace all", "replace next", and "cancel" buttons. The strings are just `Fl_Input` widgets, the "replace all" and "cancel" buttons are `Fl_Button` widgets, and the "replace next " button is a `Fl_Return_Button` widget:

Figure 4-1: The search and replace dialog.

    self.replace_dlg = Fl_Window(300, 105, "Replace")
    self.replace_find = Fl_Input(80, 10, 210, 25, "Find:")
    self.replace_with = Fl_Input(80, 40, 210, 25, "Replace:")
    self.replace_all = Fl_Button(10, 70, 90, 25, "Replace All")
    self.replace_next = Fl_Return_Button(105, 70, 120, 25, "Replace Next")
    self.replace_cancel = Fl_Button(230, 70, 60, 25, "Cancel")

## Callbacks

Now that we've defined the GUI components of our editor, we need to define our callback functions.

### changed_cb()

This function will be called whenever the user changes any text in the `editor` widget:

    def changed_cb(i1, nInserted, nDeleted, i2, c1, editor):
        global changed, loading
        if (nInserted != 0 or nDeleted != 0) and loading == False:
            changed = True
        set_title(editor);
        if loading:
            editor.editor.show_insert_position()

The `set_title()` function is one that we will write to set the changed status on the current file. We're doing it this way because we want to show the changed status in the window's title bar.

### copy_cb()

This callback function will call `kf_copy()` to copy the currently selected text to the clipboard:

    def copy_cb(widget):
        global editor
        Fl_Text_Editor.kf_copy(0, editor.editor)

### cut_cb()

This callback function will call `kf_cut()` to cut the currently selected text to the clipboard:

    def cut_cb(widget):
        global editor
        Fl_Text_Editor.kf_cut(0, editor.editor)

### delete_cb()

This callback function will call `remove_selection()` to delete the currently selected text to the clipboard:

    def delete_cb(widget):
        global textbuf
        textbuf.remove_selection()

### find_cb()

This callback function asks for a search string using the `fl_input()` convenience function and then calls the `find2_cb()` function to find the string:

    def find_cb(widget):
        global editor
        val = fl_input("Search String:", editor.search)
        if val != None:
            # User entered a string - go find it!
            editor.search = val
            find2_cb(widget)

### find2_cb()

This function will find the next occurrence of the search string. If the search string is blank then we want to pop up the search dialog:

    def find2_cb(widget):
        global editor
        if editor.search[0] == 0:
            # Search string is blank; get a new one...
            find_cb(widget, editor)
            return

        pos = editor.editor.insert_position();
        (found, pos) = textbuf.search_forward(pos, editor.search);
        if found!= 0:
            # Found a match; select and update the position...
            textbuf.select(pos, pos+len(editor.search))
            editor.editor.insert_position(pos+len(editor.search))
            editor.editor.show_insert_position()
        else:
            fl_alert("No occurrences of %s found!"%editor.search)

If the search string cannot be found we use the `fl_alert()` convenience function to display a message to that effect.

### new_cb()

This callback function will clear the editor widget and current filename. It also calls the `check_save()` function to give the user the opportunity to save the current file first as needed:

    def new_cb(widget):
        global filename, changed
        if check_save() == 0:
            return
        filename = ""
        textbuf.select(0, textbuf.length())
        textbuf.remove_selection()
        changed = False
        textbuf.call_modify_callbacks()

### open_cb()

This callback function will ask the user for a filename and then load the specified file into the input widget and current filename. It also calls the `check_save()` function to give the user the opportunity to save the current file first as needed:

    def open_cb(widget):
        global filename
        if check_save() == 0:
            return
        newfile = fl_file_chooser("Open File?", "*", filename)
        if newfile != None:
            load_file(newfile, -1)

We call the `load_file()` function to actually load the file.

### paste_cb()

This callback function will call `kf_paste()` to paste the clipboard at the current position:

    def paste_cb(widget):
        global editor
        Fl_Text_Editor.kf_paste(0, editor.editor)

### quit_cb()

The quit callback will first see if the current file has been modified, and if so give the user a chance to save it. It then exits from the program:

    def quit_cb(widget, data):
        global changed
        if changed and check_save() == 0:
            return
        sys.exit(0)

### replace_cb()

The replace callback just shows the replace dialog:

    def replace_cb(widget):
        global editor
        editor.replace_dlg.show()

### replace2_cb()

This callback will replace the next occurence of the replacement string. If nothing has been entered for the replacement string, then the replace dialog is displayed instead:

    def replace2_cb(widget):
        global editor
        find = editor.replace_find.value()
        replace = editor.replace_with.value()

        if len(find) == 0:
            editor.replace_dlg.show()
            return

        editor.replace_dlg.hide()

        pos = editor.editor.insert_position()
        (found, pos) = textbuf.search_forward(pos, find)

        if found != 0:
            # Found a match; update the position and replace text...
            textbuf.select(pos, pos+len(find))
            textbuf.remove_selection()
            textbuf.insert(pos, replace)
            textbuf.select(pos, pos+len(replace))
            editor.editor.insert_position(pos+len(replace))
            editor.editor.show_insert_position()
        else:
            fl_alert("No occurrences of %s found!"%find)

### replall_cb()

This callback will replace all occurences of the search string in the file:

    def replall_cb(widget):
        global editor
        find = editor.replace_find.value()
        replace = editor.replace_with.value()

        if len(find) == 0:
            editor.replace_dlg.show()
            return

        editor.replace_dlg.hide()
        editor.editor.insert_position(0)
        times = 0

        found = 1
        while found != 0:
            pos = editor.editor.insert_position()
            (found, pos) = textbuf.search_forward(pos, find)

            if found != 0:
                # Found a match; update the position and replace text...
                textbuf.select(pos, pos+len(find))
                textbuf.remove_selection()
                textbuf.insert(pos, replace)
                editor.editor.insert_position(pos+len(replace))
                editor.editor.show_insert_position()
                times += 1

        if times > 0:
            fl_message("Replaced %d occurrences."%times)
        else:
            fl_alert("No occurrences of %s found!"%find)

### replcan_cb()

This callback just hides the replace dialog:

    def replcan_cb(widget):
        global editor
        editor.replace_dlg.hide()

### save_cb()

This callback saves the current file. If the current filename is blank it calls the "save as" callback:

    def save_cb(widget):
        global filename
        if len(filename) == 0:
            # No filename - get one!
            saveas_cb()
            return
        else:
            save_file(filename)

The `save_file()` function saves the current file to the specified filename.

### saveas_cb()

This callback asks the user for a filename and saves the current file:

    def saveas_cb(widget, data):
        global filename
        newfile = fl_file_chooser("Save File As?", "*", filename)
        if newfile != None:
            save_file(newfile)

The `save_file()` function saves the current file to the specified filename.

## Other Functions

Now that we've defined the callback functions, we need our support functions to make it all work:

### check_save()

This function checks to see if the current file needs to be saved. If so, it asks the user if they want to save it:

    def check_save():
        global changed
        if not changed:
            return

        r = fl_choice("The current file has not been saved.\n"
                        "Would you like to save it now?",
                        "Cancel", "Save", "Don't Save")

        if r == 1:
            save_cb()
            return not changed

        if r == 2:
            return 1
        else:
            return 0
    }

### load_file()

This function loads the specified file into the `textbuf` class:

    loading = False
    def load_file(newfile, ipos):
        global changed, loading, filename
        loading = True
        if ipos != -1:
            insert = 1
            changed = True
        else:
            insert = 0
            changed = False
        if insert == 0:
            filename = ""
            r = textbuf.loadfile(newfile)
        else:
            r = textbuf.insertfile(newfile, ipos)
        if r != 0:
            fl_alert("Error reading from file %s."%newfile)
        else:
            if insert == 0:
                filename = newfile
        loading = False
        textbuf.call_modify_callbacks()

When loading the file we use the `loadfile()` method to "replace" the text in the buffer, or the `insertfile()` method to insert text in the buffer from the named file.

### save_file()

This function saves the current buffer to the specified file:

    def save_file(newfile):
        global changed, filename
        if textbuf.savefile(newfile) != 0:
            fl_alert("Error writing to file %s."%newfile)
        else:
            filename = newfile
        changed = False
        textbuf.call_modify_callbacks()

### set_title()

This function checks the `changed` variable and updates the window label accordingly:

    def set_title(win):
        global filename, title
        if len(filename) == 0:
            title = "Untitled"
        else:
            title = os.path.basename(filename)
        if changed:
            title = title+" (modified)"
        win.label(title)

## The main code

Once we've created all of the support functions, the only thing left is to tie them all together. What remains is to create a new text buffer, create a new view (window) for the text, show the window, load the file on the command-line (if any), and then enter the FLTK event loop:

    textbuf = Fl_Text_Buffer()
    style_init()

    window = new_view()
    window.show(1, sys.argv)

    if len(sys.argv) > 1:
        load_file(sys.argv[1], -1)

    Fl.run()

## Running the Editor

The complete source for our text editor can be found in the `test/editor.py` source file. You can run the program as:

    python editor.py

Congratulations, you've just built your own text editor!

## The Final Product

The final editor window should look like the image in Figure 4-2.

Figure 4-2: The completed editor window

## Advanced Features

Now that we've implemented the basic functionality, it is time to show off some of the advanced features of the `Fl_Text_Editor` widget.

### Syntax Highlighting

The `Fl_Text_Editor` widget supports highlighting of text with different fonts, colors, and sizes. The implementation is based on the excellent NEdit text editor core, which uses a parallel "style" buffer which tracks the font, color, and size of the text that is drawn.

Styles are defined using Python arrays of the form: `Fl_Text_Display::Style_Table_Entry` structure defined in `<FL/Fl_Text_Display.H>`:

    [[color, font, size], ...]

The `color` member sets the color for the text, the `font` member sets the FLTK font index to use, and the `size` member sets the pixel size of the text. The `attr` member is currently not used.

For our text editor we'll define 7 styles for plain code, comments, keywords, and preprocessor directives:

    styletable = [ # Style table
        [ FL_BLACK,      FL_COURIER,        14 ], # A - Plain
        [ FL_DARK_GREEN, FL_COURIER_ITALIC, 14 ], # B - Line comments
        [ FL_DARK_GREEN, FL_COURIER_ITALIC, 14 ], # C - Block comments
        [ FL_BLUE,       FL_COURIER,        14 ], # D - Strings
        [ FL_DARK_RED,   FL_COURIER,        14 ], # E - Directives
        [ FL_DARK_RED,   FL_COURIER_BOLD,   14 ], # F - Types
        [ FL_BLUE,       FL_COURIER_BOLD,   14 ]  # G - Keywords
    ]

You'll notice that the comments show a letter next to each style - each style in the style buffer is referenced using a character starting with the letter 'A'.

You call the `highlight_data()` method to associate the style data and buffer with the text editor widget:

    w.editor.highlight_data(stylebuf, styletable,
                                  len(styletable),
    			      'A', style_unfinished_cb, None)

Finally, you need to add a callback to the main text buffer so that changes to the text buffer are mirrored in the style buffer:

    textbuf.add_modify_callback(style_update, w.editor)

The `style_update()` function, like the `change_cb()` function described earlier, is called whenever text is added or removed from the text buffer. It mirrors the changes in the style buffer and then updates the style data as necessary:

    #
    # 'style_update()' - Update the style buffer...
    #
    def style_update(pos, nInserted, nDeleted, nRestyled, deletedText, e):
        global stylebuf, editor

        if nInserted == 0 and nDeleted == 0:
            stylebuf.unselect()
        else:
            if nInserted > 0:
                style = 'A'*nInserted
                stylebuf.replace(pos, pos+nDeleted, style)
            else:
                # Just delete characters in the style buffer...
                stylebuf.remove(pos, pos+nDeleted)

            # Select the area that was just updated to avoid unnecessary
            # callbacks...
            stylebuf.select(pos, pos + nInserted - nDeleted)
            # Re-parse the changed region; we do this by parsing from the
            # beginning of the previous line of the changed region to the end of
            # the line of the changed region...  Then we check the last
            # style character and keep updating if we have a multi-line
            # comment character...
            start = textbuf.line_start(pos)
            end   = textbuf.line_end(pos + nInserted)
            text  = textbuf.text_range(start, end)
            style = stylebuf.text_range(start, end)

            if start==end:
                last = 0
            else:
                last  = style[end - start - 1]

            style = style_parse(text, style, end - start)
            stylebuf.replace(start, end, style)
            editor.editor.redisplay_range(start, end)

            if start == end or last != style[end - start - 1]:
                # Either the user deleted some text, or the last character
                # on the line changed styles, so reparse the
                # remainder of the buffer...
                end   = textbuf.length()
                text  = textbuf.text_range(start, end)
                style = stylebuf.text_range(start, end)

                style_parse(text, style, end - start)

                stylebuf.replace(start, end, style)
                editor.editor.redisplay_range(start, end)

The `style_parse()` function scans a copy of the text in the buffer and generates the necessary style characters for display. It assumes that parsing begins at the start of a line:

    //
    // 'style_parse()' - Parse text and produce style data.
    //
    #
    # 'style_parse()' - Parse text and produce style data.
    #
    def style_parse(text, style, length):
        # Style letters:
        #
        # A - Plain
        # B - Line comments
        # C - Block comments
        # D - Strings
        # E - Directives
        # F - Types
        # G - Keywords
        col = 0
        new_style = ""
        isComment = False
        isString = False

        pos = 0
        last = False
        while pos < length:
            current = style[pos]
            if isComment:
                current = 'B'
                if text[pos] == '\n':
                    isComment = False
            elif isString:
                current = 'D'
                if text[pos] == '"' or text[pos] == "'":
                    isString = False
            elif current in ['B', 'F', 'G']:
                current = 'A'

            if current == 'A':
                # Check for directives, comments, strings, and keywords...
                if text[pos] == '#':
                    current = 'B'
                    isComment = True
                elif text[pos] == '"' or text[pos] == "'":
                    current = 'D'
                    isString = True
                elif not last and text[pos].islower() or text[pos] == '_':
                    pos1 = pos
                    while pos1 < length:
                        if text[pos1].islower() or text[pos1] == '_':
                            pos1 += 1
                        else:
                            break

                    kw = text[pos:pos1]
                    if iskeyword(kw):
                        new_style = new_style+'F'*(pos1-pos)
                        current = 'F'
                        if pos1 >= length:
                            pos1 -= 1
                        pos = pos1
                        last = True

            last = text[pos].isalnum() or text[pos] == '_' or text[pos] == '.'
            if text[pos] == '\n':
                col = 0
                if current == 'B' or current == 'E':
                    current = 'A'
            new_style = new_style+current
            pos += 1

        return  new_style